using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using System.IO;
using System.Drawing.Imaging;
using System.Drawing.Drawing2D;

namespace howto_straighten_image
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
        }

        private Bitmap OriginalBitmap = null;
        private Bitmap RotatedBitmap = null;
        private float Angle = 0;
        private Point StartPoint = new Point(50, 100);
        private Point EndPoint = new Point(210, 100);
        private bool Drawing = false;

        private void mnuFileOpen_Click(object sender, EventArgs e)
        {
            if (ofdImage.ShowDialog() == DialogResult.OK)
            {
                OriginalBitmap = LoadBitmapUnlocked(ofdImage.FileName);
                Angle = 0;
                DrawImage();
            }
        }

        private void mnuFileSaveAs_Click(object sender, EventArgs e)
        {
            if (sfdImage.ShowDialog() == DialogResult.OK)
            {
                SaveImage(RotatedBitmap, sfdImage.FileName);
            }
        }

        private void mnuFileExit_Click(object sender, EventArgs e)
        {
            Close();
        }

        // Load a bitmap without locking it.
        private Bitmap LoadBitmapUnlocked(string file_name)
        {
            using (Bitmap bm = new Bitmap(file_name))
            {
                return new Bitmap(bm);
            }
        }

        // Save the file with the appropriate format.
        public void SaveImage(Image image, string filename)
        {
            string extension = Path.GetExtension(filename);
            switch (extension.ToLower())
            {
                case ".bmp":
                    image.Save(filename, ImageFormat.Bmp);
                    break;
                case ".exif":
                    image.Save(filename, ImageFormat.Exif);
                    break;
                case ".gif":
                    image.Save(filename, ImageFormat.Gif);
                    break;
                case ".jpg":
                case ".jpeg":
                    image.Save(filename, ImageFormat.Jpeg);
                    break;
                case ".png":
                    image.Save(filename, ImageFormat.Png);
                    break;
                case ".tif":
                case ".tiff":
                    image.Save(filename, ImageFormat.Tiff);
                    break;
                default:
                    throw new NotSupportedException(
                        "Unknown file extension " + extension);
            }
        }

        // Display the image rotated by the current amount.
        private void DrawImage()
        {
            if (OriginalBitmap == null)
            {
                picImage.Refresh();
                return;
            }

            // Get the image's dimensions.
            int wid = OriginalBitmap.Width;
            int hgt = OriginalBitmap.Height;

            // Make a Matrix representing the rotation.
            Matrix matrix = new Matrix();
            matrix.Rotate(Angle);

            // Rotate the image's corners to see
            // how large the rotated image must be.
            PointF[] points =
            {
                new PointF(0, 0),
                new PointF(wid, 0),
                new PointF(0, hgt),
                new PointF(wid, hgt),
            };
            matrix.TransformPoints(points);

            // Get the rotated bounds.
            float xmin = points[0].X;
            float xmax = xmin;
            float ymin = points[0].Y;
            float ymax = ymin;
            for (int i = 1; i < points.Length; i++)
            {
                if (xmin > points[i].X) xmin = points[i].X;
                if (xmax < points[i].X) xmax = points[i].X;
                if (ymin > points[i].Y) ymin = points[i].Y;
                if (ymax < points[i].Y) ymax = points[i].Y;
            }

            // Get the new image's dimensions.
            float new_wid = xmax - xmin;
            float new_hgt = ymax - ymin;

            // Add a translation to move the rotated image
            // to the center of the new bitmap.
            matrix.Translate(new_wid / 2, new_hgt / 2, MatrixOrder.Append);

            // Make the new bitmap.
            RotatedBitmap = new Bitmap((int)new_wid, (int)new_hgt);
            using (Graphics gr = Graphics.FromImage(RotatedBitmap))
            {
                gr.InterpolationMode = InterpolationMode.High;
                gr.Clear(Color.White);
                gr.Transform = matrix;

                // Draw the image centered at the origin.
                PointF[] dest_points =
                {
                    new PointF(-wid / 2, -hgt / 2),
                    new PointF(wid / 2, -hgt / 2),
                    new PointF(-wid / 2, hgt / 2),
                };
                gr.DrawImage(OriginalBitmap, dest_points);
            }

            // Display the result.
            picImage.Image = RotatedBitmap;
        }

        // Let the use select an orientation line.
        private void picImage_MouseDown(object sender, MouseEventArgs e)
        {
            StartPoint = e.Location;
            EndPoint = e.Location;
            Drawing = true;
        }

        private void picImage_MouseMove(object sender, MouseEventArgs e)
        {
            if (!Drawing) return;
            EndPoint = e.Location;
            picImage.Refresh();
        }

        private void picImage_MouseUp(object sender, MouseEventArgs e)
        {
            Drawing = false;

            // Calculate the line's angle.
            int dx = EndPoint.X - StartPoint.X;
            int dy = EndPoint.Y - StartPoint.Y;
            double new_angle = Math.Atan2(dy, dx) * 180.0 / Math.PI;
            
            // Subtract this angle from the total angle so far.
            Angle -= (float)new_angle;

            // Make the new rotated image.
            DrawImage();
        }

        // Draw the horizon line and alignment grid.
        private void picImage_Paint(object sender, PaintEventArgs e)
        {
            if (Drawing)
            {
                e.Graphics.SmoothingMode = SmoothingMode.AntiAlias;
                e.Graphics.DrawLine(Pens.Yellow, StartPoint, EndPoint);
                using (Pen pen = new Pen(Color.Red))
                {
                    pen.DashPattern = new float[] { 5, 5 };
                    e.Graphics.DrawLine(pen, StartPoint, EndPoint);
                }
            }

            if (mnuFormatDrawGrid.Checked)
            {
                const int dx = 50;
                int xmax = picImage.ClientSize.Width;
                int ymax = picImage.ClientSize.Height;
                for (int x = 0; x < xmax; x += dx)
                    e.Graphics.DrawLine(Pens.Silver, x, 0, x, ymax);
                for (int y = 0; y < ymax; y += dx)
                    e.Graphics.DrawLine(Pens.Silver, 0, y, xmax, y);
            }
        }

        // Reset the angle to 0.
        private void mnuFormatReset_Click(object sender, EventArgs e)
        {
            Angle = 0;
            StartPoint = new Point(50, 100);
            EndPoint = new Point(210, 100);
            DrawImage();
        }

        private void mnuFormatDrawGrid_Click(object sender, EventArgs e)
        {
            picImage.Refresh();
        }
    }
}
